import unittest as ut
from spvcm import utils
from spvcm._constants import RTOL, ATOL, TEST_SEED, CLASSTYPES
from warnings import warn
import numpy as np
import copy
import types

class Model_Mixin(object):
    def build_self(self):
        super(Model_Mixin, self).__init__()
        self.inputs = utils.south()
        self.__dict__.update(self.inputs)
        self.ignore_shape = False
        self.squeeze = True

    def test_trace(self):
        self.inputs['n_samples'] = 0
        instance = self.cls(**self.inputs)
        np.random.seed(TEST_SEED)
        instance.draw()
        instance.trace._assert_allclose(self.answer_trace,
                                        rtol=RTOL, atol=ATOL,
                                        ignore_shape = self.ignore_shape,
                                        squeeze=self.squeeze)

    @ut.skip
    def test_argument_parsing(self):
        #priors, initial values, etc.
        raise NotImplementedError

    @ut.skip
    def test_covariance_assignment(self):
        raise NotImplementedError

def run_with_seed(cls, env=utils.south(), seed=TEST_SEED, fprefix = ''):
    fname = str(cls).strip("'<>'").split('.')[-1].lower()
    try:
        env['n_samples'] = 0
        model = cls(**env)
    except TypeError:
        reduced = copy.deepcopy(env)
        del reduced['M']
        del reduced['W']
        reduced['n_samples'] = 0
        model = cls(**reduced)
    np.random.seed(TEST_SEED)
    model.draw()
    model.trace.to_df().to_csv(fprefix + fname + '.csv', index=False)



from __future__ import division
from scipy import sparse as spar
import numpy as np
from numpy import linalg as nla
from scipy.sparse import linalg as spla
import libpysal
import scipy.linalg as scla
from warnings import warn as Warn

__all__ = ['grid_det']
PUBLIC_DICT_ATTS = [k for k in dir(dict) if not k.startswith('_')]

##########################
# GENERAL PURPOSE UTILS  #
##########################

def no_op(*_, **__):
    """
    This is a no-op. It takes any arguments,
    keyword or positional, and returns none
    """
    return

def thru_op(*args, **kws):
    """
    This is a thru-op. It returns everything passed to it.
    """
    if len(args) == 1 and kws == dict():
        return args[0]
    elif len(args) == 1 and kws != dict():
        return args[0], kws
    elif len(args) == 0:
        return kws
    else:
        return args, kws

def zero_op(*args, **kws):
    """
    this is a zero op. It takes any arguments, keyword
    or positional, and returns zero.
    """
    return 0

##########################
# BUILD EXAMPLE DATASETS #
##########################

def south(df=False):
    """
    Sets up the data for the US southern counties example.

    Returns
    -------
    dictionary or (dictionary, dataframe), where the dictionary is keyed on:

    X           : Data from southern counties, columns "GI89", "BLK90", "HR90"
    Y           : Outcome variate, "DNL90"
    Z           : upper-level variate, the state average "FH90"
    W           : queen weights matrix between counties
    M           : queen matrix between states
    membership  : membership vector relating counties to their states

    and the dataframe contains the raw dataset
    """
    import geopandas
    south = libpysal.examples.load_example('South')
    data = geopandas.read_file(south.get_path('south.shp'))
    data = data[data.STATE_NAME != 'District of Columbia']
    X = data[['GI89', 'BLK90', 'HR90']].values
    N = X.shape[0]
    Z = data.groupby('STATE_NAME')['FH90'].mean()
    Z = Z.values.reshape(-1,1)
    J = Z.shape[0]

    Y = data.DNL90.values.reshape(-1,1)
    
    us48 = libpysal.examples.load_example('us_income')
    W2 = libpysal.weights.Queen.from_shapefile(us48.get_path('us48.shp'),
                                               idVariable='STATE_NAME')
    W2 = libpysal.weights.set_operations.w_subset(W2, ids=data.STATE_NAME.unique().tolist()) #only keep what's in the data
    W1 = libpysal.weights.Queen.from_shapefile(south.get_path('south.shp'),
                                            idVariable='FIPS')
    W1 = libpysal.weights.set_operations.w_subset(W1, ids=data.FIPS.tolist()) #again, only keep what's in the data

    W1.transform = 'r'
    W2.transform = 'r'

    membership = data.STATE_NAME.apply(lambda x: W2.id_order.index(x)).values

    d = {'X':X, 'Y':Y, 'Z':Z, 'W':W1, 'M':W2, 'membership':membership}
    if df:
        return d, data
    else:
        return d

def baltim(df=False):
    """
    Sets up the baltimore house price example

    Returns
    --------
    dictionary or (dictinoary, dataframe), where the dictionary is keyed:

    X           : Data from baltimore houses, columns "AGE", "LOTSZ", "SQFT"
    Y           : outcomes, log house price
    coordinates : the geographic coordinates of house sales

    dataframe contains the raw data of the baltimore example
    """
    import geopandas
    baltim = libpysal.examples.load_example('Baltimore')
    baltim = geopandas.read_file(baltim.get_path('baltim.shp'))
    coords = baltim[['X', 'Y']].values
    Y = np.log(baltim.PRICE.values).reshape(-1,1)
    Yz = Y - Y.mean()
    X = baltim[['AGE', 'LOTSZ', 'SQFT']].values
    Xz = X-X.mean(axis=0)
    out = {'Y':Yz, 'X':Xz, 'coordinates':coords}
    if df:
        return out, baltim
    else:
        return out

####################
# MATRIX UTILITIES #
####################

def lulogdet(matrix):
    """
    compute the log determinant using a lu decomposition appropriate to input type
    """
    if spar.issparse(matrix):
        LUfunction = lambda x: spla.splu(x).U.diagonal()
    else:
        LUfunction = lambda x: scla.lu_factor(x)[0].diagonal()
    LUdiag = LUfunction(matrix)
    return np.sum(np.log(np.abs(LUdiag)))

def splogdet(matrix):
    """
    compute the log determinant via an appropriate method according to the input.
    """
    redo = False
    if spar.issparse(matrix):
        LU = spla.splu(spar.csc_matrix(matrix))
        ldet = np.sum(np.log(np.abs(LU.U.diagonal())))
    else:
        sgn, ldet = nla.slogdet(matrix)
        if np.isinf(ldet) or sgn is 0:
            Warn('Dense log determinant via numpy.linalg.slogdet() failed!')
            redo = True
        if sgn not in [-1,1]:
            Warn("Drastic loss of precision in numpy.linalg.slogdet()!")
            redo = True
        ldet = sgn*ldet
    if redo:
        Warn("Please pass convert to a sparse weights matrix. Trying sparse determinant...", UserWarning)
        ldet = splogdet(spar.csc_matrix(matrix))
    return ldet

def speye(i, sparse=True):
    """
    constructs a square identity matrix according to i, either sparse or dense
    """
    if sparse:
        return spar.identity(i)
    else:
        return np.identity(i)

spidentity = speye

def speye_like(matrix):
    """
    constructs an identity matrix depending on the input dimension and type
    """
    if matrix.shape[0] != matrix.shape[1]:
        raise UserWarning("Matrix is not square")
    else:
        return speye(matrix.shape[0], sparse=spar.issparse(matrix))

spidentity_like = speye_like

def speigen_range(matrix, retry=True, coerce=True):
    """
    Construct the eigenrange of a potentially sparse matrix.
    """
    if spar.issparse(matrix):
        try:
            emax = spla.eigs(matrix, k=1, which='LR')[0]
        except (spla.ArpackNoConvergence, spla.ArpackError) as e:
            rowsums = np.unique(np.asarray(matrix.sum(axis=1)).flatten())
            if np.allclose(rowsums, np.ones_like(rowsums)):
                emax = np.array([1])
            else:
                Warn('Maximal eigenvalue computation failed to converge'
                     ' and matrix is not row-standardized.')
                raise e
        emin = spla.eigs(matrix, k=1, which='SR')[0]
        if coerce:
            emax = emax.real.astype(float)
            emin = emin.real.astype(float)
    else:
        try:
            eigs = nla.eigvals(matrix)
            emin, emax = eigs.min().astype(float), eigs.max().astype(float)
        except Exception as e:
            Warn('Dense eigenvector computation failed!')
            if retry:
                Warn('Retrying with sparse matrix...')
                spmatrix = spar.csc_matrix(matrix)
                speigen_range(spmatrix)
            else:
                Warn('Bailing...')
                raise e
    return emin, emax

def spinv(M):
    """
    Compute an inverse of a matrix using the appropriate sparse or dense
    function
    """
    if spar.issparse(M):
        return spla.inv(M)
    else:
        return nla.inv(M)

def spsolve(A,b):
    """
    Solve the system Ax=b for x, depending on the type of A. The solution vector is equivalent to A^{-1}b

    If a is sparse, the result will be sparse. Otherwise, the result will be dense.
    """
    if spar.issparse(A):
        return spla.spsolve(A, b)
    elif spar.issparse(b):
        Warn('b is sparse, but A is dense. Solving the dense system.')
        return spsolve(A, b.toarray())
    return scla.solve(A,b)



#########################
# STATISTICAL UTILITIES #
#########################

def chol_mvn(Mu, Sigma):
    """
    Sample from a Multivariate Normal given a mean & Covariance matrix, using
    cholesky decomposition of the covariance. If the cholesky decomp fails due
    to the matrix not being strictly positive definite, then the
    numpy.random.multivariate_normal will be used.

    That is, new values are generated according to :
    New = Mu + cholesky(Sigma) . N(0,1)

    Parameters
    ----------
    Mu      :   np.ndarray (p,1)
                An array containing the means of the multivariate normal being
                sampled
    Sigma   :   np.ndarray (p,p)
                An array containing the covariance between the dimensions of the
                multivariate normal being sampled

    Returns
    -------
    np.ndarray of size (p,1) containing draws from the multivariate normal
    described by MVN(Mu, Sigma)
    """
    try:
        D = scla.cholesky(Sigma, overwrite_a = True)
        e = np.random.normal(0,1,size=Mu.shape)
        kernel = np.dot(D.T, e)
        out = Mu + kernel
    except np.linalg.LinAlgError:
        out = np.random.multivariate_normal(Mu.flatten(), Sigma)
        out = out.reshape(Mu.shape)
    return out

def sma_covariance(param, W, sparse=True):
    """
    This computes a covariance matrix for a SMA-type error specification:

    ( (I + param * W)(I + param * W)^T)

    this always returns a dense array
    """
    half = speye_like(W) + param * W
    whole = half.dot(half.T)
    if sparse:
        return whole
    return whole.toarray()

def sma_precision(param, W, sparse=False):
    """
    This computes a precision matrix for a spatial moving average error specification.
    """
    covariance = sma_covariance(param, W, sparse=sparse)
    if sparse:
        return spinv(covariance)
    return np.linalg.inv(covariance)

def se_covariance(param, W, sparse=False):
    """
    This computes a covariance matrix for a SAR-type error specification:

    ( (I - param * W)^T(I - param * W) )^{-1}

    and always returns a dense matrix.

    This first calls se_precision, and then inverts the results of that call.

    """
    prec = se_precision(param, W, sparse=sparse)
    if sparse:
        return spla.inv(prec)
    return np.linalg.inv(prec)

def se_precision(param, W, sparse=True):
    """
    This computes a precision matrix for a SAR-type error specification.
    """
    half = speye_like(W) - param * W
    prec = half.T.dot(half)
    if sparse:
        return prec
    return prec.toarray()

def ind_covariance(param, W, sparse=False):
    """
    This returns a covariance matrix for a standard diagonal specification:

    I

    and always returns a dense matrix. Thus, it ignores param entirely.
    """
    out = speye(W.shape[0], sparse=sparse)
    if sparse:
        return spar.csc_matrix(out)
    return out

def grid_det(W, parmin=None, parmax=None, parstep=None, grid=None):
    """
    This is a utility function to set up the grid of matrix determinants over a
    range of spatial parameters for a given W.
    """
    if (parmin is None) and (parmax is None):
        parmin, parmax = speigen_range(W)
    if parstep is None:
        parstep = (parmax - parmin) / 1000
    if grid is None:
        grid = np.arange(parmin, parmax, parstep)
    logdets = [splogdet(speye_like(W) - rho * W) for rho in grid]
    grid = np.vstack((grid, np.array(logdets).reshape(grid.shape)))
    return grid



import sys
import types
import os
RTOL = 1e-3
ATOL = 1e-3
TEST_SEED = 310516

PY2 = sys.version_info[0] == 2

if PY2:
    CLASSTYPES = (types.ClassType, type)
else:
    CLASSTYPES = (type,)

PACKAGE_PATH = os.path.dirname(os.path.abspath(__file__))



import warnings
from datetime import datetime as dt
import numpy as np
import copy
import multiprocessing as mp
import pandas as pd
import os

from .sqlite import head_to_sql, start_sql
from .plotting import plot_trace
from collections import OrderedDict

import six
if not six.PY3:
    range = xrange


__all__ = ['Sampler_Mixin', 'Hashmap', 'Trace']

######################
# SAMPLER MECHANISMS #
######################


class Sampler_Mixin(object):
    """
    A Mixin class designed to facilitate code reuse. This should be the parent class of anything that uses the sampling framework in this package.
    """
    def __init__(self):
        super(Sampler_Mixin, self).__init__()

    def sample(self, n_samples, n_jobs=1):
        """
        Sample from the joint posterior distribution defined by all of the
        parameters in the gibbs sampler.

        Parameters
        ----------
        n_samples   :   int
                        number of samples from the joint posterior density to take
        n_jobs      :   int
                        number of parallel chains to run.

        Returns
        -------
        Implicitly updates all values in place, returns None
        """
        try:
            from tqdm import tqdm
        except ImportError:
            from .utils import thru_op
            tqdm = thru_op
            msg = '`tqdm` is not available. '
            msg += 'Using `spvcm.utils.thru_op` in place of `tqdm`.'
            warnings.warn(msg, stacklevel=2)
            
        if n_jobs > 1:
           self._parallel_sample(n_samples, n_jobs)
           return
        elif isinstance(self.state, list):
            self._parallel_sample(n_samples, n_jobs=len(self.state))
            return
        _start = dt.now()
        try:
            for _ in tqdm(range(n_samples)):
                if (self._verbose > 1) and (n_samples % 100 == 0):
                    print('{} Draws to go'.format(n_samples))
                self.draw()
        except KeyboardInterrupt:
            warnings.warn('Sampling interrupted, drew {} samples'.format(self.cycles))
        finally:
            _stop = dt.now()
            if not hasattr(self, 'total_sample_time'):
                self.total_sample_time = _stop - _start
            else:
                self.total_sample_time += _stop - _start

    def draw(self):
        """
        Take exactly one sample from the joint posterior distribution.
        """
        if self.cycles == 0:
            self._finalize()
        self._iteration()
        self.cycles += 1
        for param in self.traced_params:
            self.trace.chains[0][param].append(self.state[param])
        if self.database is not None:
            head_to_sql(self, self._cur, self._cxn)
            for param in self.traced_params:
                self.trace.chains[0][param] = [self.trace[param,-1]]

    def _parallel_sample(self, n_samples, n_jobs):
        """
        Run n_jobs parallel samples of a given model. 
        Not intended to be called directly, and should be called by model.sample.
        """
        models = [copy.deepcopy(self) for _ in range(n_jobs)]
        for i, model in enumerate(models):
            if isinstance(model.state, list):
                models[i].state = copy.deepcopy(self.state[i])
            if hasattr(model, 'configs'):
                if isinstance(model.configs, list):
                    models[i].configs = copy.deepcopy(self.configs[i])
            if self.database is not None:
                models[i].database = self.database + str(i)
            models[i].trace = Trace(**{k:[] for k in model.trace.varnames})
            if self.cycles == 0:
                models[i]._fuzz_starting_values()
        n_samples = [n_samples] * n_jobs
        _start = dt.now()
        seed = np.random.randint(0,10000, size=n_jobs).tolist()
        P = mp.Pool(n_jobs)
        results = P.map(_reflexive_sample, zip(models, n_samples, seed))
        P.close()
        _stop = dt.now()
        if self.cycles > 0:
            new_traces = []
            for i, model in enumerate(results):
                # model.trace.chains is always single-chain, since we've broken everything into single chains
                new_traces.append(Hashmap(**{k:param + model.trace.chains[0][k]
                                             for k, param in self.trace.chains[i].items()}))
            new_trace = Trace(*new_traces)
        else:
            new_trace = Trace(*[model.trace.chains[0] for model in results])
        self.trace = new_trace
        self.state = [model.state for model in results]
        self.cycles += n_samples[0]
        self.configs = [model.configs for model in results]
        if hasattr(self, 'total_sample_time'):
            self.total_sample_time += _stop - _start
        else:
            self.total_sample_time = _stop - _start

    def _fuzz_starting_values(self, state=None):
        """
        Function to overdisperse starting values used in the package.
        """
        st = self.state
        if hasattr(st, 'Betas'):
            st.Betas += np.random.normal(0,5, size=st.Betas.shape)
        if hasattr(st, 'Alphas'):
            st.Alphas += np.random.normal(0,5,size=st.Alphas.shape)
        if hasattr(st, 'Sigma2'):
            st.Sigma2 += np.random.uniform(0,5)
        if hasattr(st, 'Tau2'):
            st.Tau2 += np.random.uniform(0,5)
        if hasattr(st, 'Lambda'):
            st.Lambda += np.random.uniform(-.25,.25)
        if hasattr(st, 'Rho'):
            st.Rho += np.random.uniform(-.25,.25)

    def _finalize(self, **args):
        """
        Abstract function to ensure inheritors define a finalze method. This method should compute all derived quantities used in the _iteration() function that would change if the user changed priors, starting values, or other information. This is to ensure that if the user initializes the sampler with n_samples=0 and then changes the state, the derived quantites used in sampling are correct.
        """
        raise NotImplementedError

    def _setup_priors(self, **args):
        """
        Abstract function to ensure inheritors define a _setup_priors method. This method should assign into the state all of the correct priors for all parameters in the model.
        """
        raise NotImplementedError

    def _setup_truncation(self, **args):
        """
        Abstract function to ensure inheritors define a _setup_truncation method. This method should truncate parameter space to a given arbitrary bounds.
        """
        raise NotImplementedError

    def _setup_starting_values(self, **args):
        """
        Abstract function to ensure that inheritors define a _setup_starting_values method. This method should assign the correct values for each of the parameters into model.state.
        """
        raise NotImplementedError

    @property
    def database(self):
        """
        the database used for the model.
        """
        return getattr(self, '_db', None)

    @database.setter
    def database(self, filename):
        self._cxn, self._cur = start_sql(self, tracename=filename)
        self._db = filename
        from .sqlite import trace_from_sql
        def load_sqlite():
            return trace_from_sql(filename)
        self.trace.load_sqlite = load_from_sqlite

def _reflexive_sample(tup):
    """
    a helper function sample a bunch of models in parallel.

    Tuple must be:

    model : model object
    n_samples : int number of samples
    seed : seed to use for the sampler
    """
    model, n_samples, seed = tup
    np.random.seed(seed)
    model.sample(n_samples=n_samples)
    return model

def _noop(*args, **kwargs):
    pass

#######################
# MAPS AND CONTAINERS #
#######################

class Hashmap(dict):
    """
    A dictionary with dot access on attributes
    """
    def __init__(self, **kw):
        super(Hashmap, self).__init__(**kw)
        if kw != dict():
            for k in kw:
                self[k] = kw[k]

    def __getattr__(self, attr):
        try:
            r = self[attr]
        except KeyError:
            try:
                r = getattr(super(Hashmap, self), attr)
            except AttributeError:
                raise AttributeError("'{}' object has no attribute '{}'"
                                     .format(self.__class__, attr))
        return r

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(Hashmap, self).__setitem__(key,value)
        self.__dict__.update({key:value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Hashmap, self).__delitem__(key)
        del self.__dict__[key]

class Trace(object):
    """
    Object to contain results from sampling.

    Arguments
    ---------
    chains  :   a chain or comma-separated sequence of chains
                a chain is a dict-like collection, where keys are the parameter name and the values are the values of the chain.
    kwargs  :   a dictionary splatted into keyword arguments
                the name of the argument is taken to the be the parameter name, and the value is taken to be a chain of that parameter.

    Examples
    ---------
    >>> Trace(a=[1,2,3], b=[4,2,5], c=[1,9,23]) #Trace with one chain
    >>> Trace([{'a':[1,2,3], 'b':[4,2,5], 'c':[1,9,23]},
               {'a':[2,5,1], 'b':[2,9,1], 'c':[9,21,1]}]) #Trace with two chains
    """
    def __init__(self, *chains, **kwargs):
        if chains is () and kwargs != dict():
            self.chains = _maybe_hashmap(kwargs)
        if chains is not ():
            self.chains = _maybe_hashmap(*chains)
            if kwargs != dict():
                self.chains.extend(_maybe_hashmap(kwargs))
        self._validate_schema()

    @property
    def varnames(self, chain=None):
        """
        Names of variables contained in the trace.
        """
        try:
            return self._varnames
        except AttributeError:
            try:
                self._validate_schema()
            except KeyError:
                if chain is None:
                    raise Exception('Variable names are heterogeneous in chains and no default index provided.')
                else:
                    warnings.warn('Variable names are heterogeneous in chains!', stacklevel=2)
                    return list(self.chains[chain].keys())
            self._varnames = list(self.chains[0].keys())
            return self._varnames

    def drop(self, varnames, inplace=True):
        """
        Drop a variable from the trace.

        Arguments
        ---------
        varnames    :   list of strings
                        names of parameters to drop from the trace.
        inplace     :   bool
                        whether to return a copy of the trace with parameters removed, or remove them inplace.
        """
        if isinstance(varnames, str):
            varnames = (varnames,)
        if not inplace:
            new = copy.deepcopy(self)
            new.drop(varnames, inplace=True)
            new._varnames = list(new.chains[0].keys())
            return new
        for i, chain in enumerate(self.chains):
            for varname in varnames:
                del self.chains[i][varname]
        self._varnames = list(self.chains[0].keys())

    def _validate_schema(self, chains=None):
        """
        Validates the trace to ensure that the chain is self-consistent.
        """
        if chains is None:
            chains = self.chains
        tracked_in_each = [set(chain.keys()) for chain in chains]
        same_schema = [names == tracked_in_each[0] for names in tracked_in_each]
        try:
            assert all(same_schema)
        except AssertionError:
            bad_chains = [i for i in range(len(chains)) if same_schema[i]]
            KeyError('The parameters tracked in each chain are not the same!'
                     '\nChains {} do not have the same parameters as chain 1!'.format(bad_chains))

    def add_chain(self, chains, validate=True):
        """
        Add chains to a trace object

        Parameters
        ----------
        chains  :   Hashmap or list of hashmaps
                    chains to merge into the trace
        validate:   bool
                    whether or not to validate the schema and reject the chain if it does not match the current trace.
        """
        if not isinstance(chains, (list, tuple)):
            chains = (chains,)
        new_chains = [self.chains]
        for chain in chains:
            if isinstance(chain, Hashmap):
                new_chains.append(chain)
            elif isinstance(chain, Trace):
                new_chains.extend(chain.chains)
            else:
                new_chains.extend(_maybe_hashmap(chain))
        if validate:
            self._validate_schema(chains=new_chains)
        self.chains = new_chains

    def map(self, func, **func_args):
        """
        Map a function over all parameters in a chain.
        Multivariate parameters are reduced to sequences of univariate parameters.

        Usage
        -------
        Intended when full-trace statistics are required. Most often,
        the trace should be sliced directly. For example, to get the mean value of a
        parameter over the last -1000 iterations with a thinning of 2:

        trace[0, 'Betas', -1000::2].mean(axis=0)

        but, to average of the parameter over all recorded chains:

        trace['Betas', -1000::2].mean(axis=0).mean(axis=0)

        since the first reduction provides an array where rows
        are iterations and columns are parameters.

        trace.map(np.mean) yields the mean of each parameter within each chain, and is
        provided to make within-chain reductions easier.

        Arguments
        ---------
        func        :   callable
                        a function that returns a result when provided a flat vector.
        varnames    :   string or list of strings
                        a keyword only argument governing which parameters to map over.
        func_args   :   dictionary/keyword arguments
                        arguments needed to be passed to the reduction
        """
        varnames = func_args.pop('varnames', self.varnames)
        if isinstance(varnames, str):
            varnames = (varnames, )
        all_stats = []
        for i, chain in enumerate(self.chains):
            these_stats=dict()
            for var in varnames:
                data = np.squeeze(self[i,var])
                if data.ndim > 1:
                    n,p = data.shape[0:2]
                    rest = data.shape[2:0]
                    if len(rest) == 0:
                        data = data.T
                    elif len(rest) == 1:
                        data = data.reshape(n,p*rest[0]).T
                    else:
                        raise Exception('Parameter "{}" shape not understood.'                  ' Please extract, shape it, and pass '
                                        ' as its own chain. '.format(var))
                else:
                    data = data.reshape(1,-1)
                stats = [func(datum, **func_args) for datum in data]
                if len(stats) == 1:
                    stats = stats[0]
                these_stats.update({var:stats})
            all_stats.append(these_stats)
        return all_stats

    @property
    def n_chains(self):
        return len(self.chains)

    @property
    def n_iters(self):
        """
        Number of raw iterations stored in the trace.
        """
        lengths = [len(chain[self.varnames[0]]) for chain in self.chains]
        if len(lengths) == 1:
            return lengths[0]
        else:
            return lengths

    def plot(self, burn=0, thin=None, varnames=None,
             kde_kwargs={}, trace_kwargs={}, figure_kwargs={}):
        """
        Make a trace plot paired with a distributional plot.

        Arguments
        -----------
        trace   :   namespace
                    a namespace whose variables are contained in varnames
        burn    :   int
                    the number of iterations to discard from the front of the trace
        thin    :   int
                    the number of iterations to discard between iterations
        varnames :  str or list
                    name or list of names to plot.
        kde_kwargs : dictionary
                     dictionary of aesthetic arguments for the kde plot
        trace_kwargs : dictionary
                       dictinoary of aesthetic arguments for the traceplot

        Returns
        -------
        figure, axis tuple, where axis is (len(varnames), 2)
        """
        f, ax = plot_trace(model=None, trace=self, burn=burn,
                           thin=thin, varnames=varnames,
                      kde_kwargs=kde_kwargs, trace_kwargs=trace_kwargs,
                      figure_kwargs=figure_kwargs)
        return f,ax

    def summarize(self, level=0):
        """
        Compute a summary of the trace. See Also: diagnostics.summary

        Arguments
        ------------
        level   :   int
                    0 for a summary by chain or 1 if the summary should be computed by pooling over chains.
        """
        from .diagnostics import summarize
        return summarize(trace=self, level=level)

    def __getitem__(self, key):
        """
        Getting an item from a trace can be done using at most three indices, where:

        1 index
        --------
            str/list of str: names of variates in all chains to grab. Returns list of Hashmaps
            slice/int: iterations to grab from all chains. Returns list of Hashmaps, sliced to the specification

        2 index
        -------
            (str/list of str, slice/int): first term is name(s) of variates in all chains to grab,
                                          second term specifies the slice each chain.
                                          returns: list of hashmaps with keys of first term and entries sliced by the second term.
            (slice/int, str/list of str): first term specifies which chains to retrieve,
                                          second term is name(s) of variates in those chains
                                          returns: list of hashmaps containing all iterations
            (slice/int, slice/int): first term specifies which chains to retrieve,
                                    second term specifies the slice of each chain.
                                    returns: list of hashmaps with entries sliced by the second term
        3 index
        --------
            (slice/int, str/list of str, slice/int) : first term specifies which chains to retrieve,
                                                      second term is the name(s) of variates in those chains,
                                                      third term is the iteration slicing.
                                                      returns: list of hashmaps keyed on second term, with entries sliced by the third term
        """
        if isinstance(key, str): #user wants only one name from the trace
            if self.n_chains  > 1:
                result = ([chain[key] for chain in self.chains])
            else:
                result = (self.chains[0][key])
        elif isinstance(key, (slice, int)): #user wants all draws past a certain index
            if self.n_chains > 1:
                return [Hashmap(**{k:v[key] for k,v in chain.items()}) for chain in self.chains]
            else:
                return Hashmap(**{k:v[key] for k,v in self.chains[0].items()})
        elif isinstance(key, list) and all([isinstance(val, str) for val in key]): #list of atts over all iters and all chains
                if self.n_chains > 1:
                    return [Hashmap(**{k:chain[k] for k in key}) for chain in self.chains]
                else:
                    return Hashmap(**{k:self.chains[0][k] for k in key})
        elif isinstance(key, tuple): #complex slicing
            if len(key) == 1:
                return self[key[0]] #ignore empty blocks
            if len(key) == 2:
                head, tail = key
                if isinstance(head, str): #all chains, one var, some iters
                    if self.n_chains > 1:
                        result = ([_ifilter(tail, chain[head]) for chain in self.chains])
                    else:
                        result = (_ifilter(tail, self.chains[0][head]))
                elif isinstance(head, list) and all([isinstance(v, str) for v in head]): #all chains, some vars, some iters
                    if self.n_chains > 1:
                        return [Hashmap(**{name:_ifilter(tail, chain[name]) for name in head})
                                   for chain in self.chains]
                    else:
                        chain = self.chains[0]
                        return Hashmap(**{name:_ifilter(tail, chain[name]) for name in head})
                elif isinstance(tail, str):
                    target_chains = _ifilter(head, self.chains)
                    if isinstance(target_chains, Hashmap):
                        target_chains = [target_chains]
                    if len(target_chains) > 1:
                        result = ([chain[tail] for chain in target_chains])
                    elif len(target_chains) == 1:
                        result = (target_chains[0][tail])
                    else:
                        raise IndexError('The supplied chain index {} does not'
                                        ' match any chains in trace.chains'.format(head))
                elif isinstance(tail, list) and all([isinstance(v, str) for v in tail]):
                    target_chains = _ifilter(head, self.chains)
                    if isinstance(target_chains, Hashmap):
                        target_chains = [target_chains]
                    if len(target_chains) > 1:
                        return [Hashmap(**{k:chain[k] for k in tail}) for chain in target_chains]
                    elif len(target_chains) == 1:
                        return Hashmap(**{k:target_chains[0][k] for k in tail})
                    else:
                        raise IndexError('The supplied chain index {} does not'
                                         ' match any chains in trace.chains'.format(head))
                else:
                    target_chains = _ifilter(head, self.chains)
                    if isinstance(target_chains, Hashmap):
                        target_chains = [target_chains]
                    out = [Hashmap(**{k:_ifilter(tail, val) for k,val in chain.items()})
                            for chain in target_chains]
                    if len(out) == 1:
                        return out[0]
                    else:
                        return out
            elif len(key) == 3:
                chidx, varnames, iters = key
                if isinstance(chidx, int):
                    if np.abs(chidx) > self.n_chains:
                        raise IndexError('The supplied chain index {} does not'
                                         ' match any chains in trace.chains'.format(chidx))
                if varnames == slice(None, None, None):
                    varnames = self.varnames
                chains = _ifilter(chidx, self.chains)
                if isinstance(chains, Hashmap):
                    chains = [chains]
                nchains = len(chains)
                if isinstance(varnames, str):
                    varnames = [varnames]
                if varnames is slice(None, None, None):
                    varnames = self.varnames
                if len(varnames) == 1:
                    if nchains > 1:
                        result = ([_ifilter(iters, chain[varnames[0]]) for chain in chains])
                    else:
                        result = (_ifilter(iters, chains[0][varnames[0]]))
                else:
                    if nchains > 1:
                        return [Hashmap(**{varname:_ifilter(iters, chain[varname])
                                        for varname in varnames})
                                for chain in chains]
                    else:
                        return Hashmap(**{varname:_ifilter(iters, chains[0][varname]) for varname in varnames})
        else:
            raise IndexError('index not understood')

        result = np.asarray(result)
        if result.shape == ():
            result = result.tolist()
        elif result.shape in [(1,1), (1,)]:
            result = result[0]
        return result

    ##############
    # Comparison #
    ##############

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            a = [ch1==ch2 for ch1,ch2 in zip(other.chains, self.chains)]
            return all(a)

    def _allclose(self, other, **allclose_kw):
        try:
            self._assert_allclose(other, **allclose_kw)
        except AssertionError:
            return False
        return True

    def _assert_allclose(self, other, **allclose_kw):
        ignore_shape = allclose_kw.pop('ignore_shape', False)
        squeeze = allclose_kw.pop('squeeze', True)
        try:
            assert set(self.varnames) == set(other.varnames)
        except AssertionError:
            raise AssertionError('Variable names are different!\n'
                                 'self: {}\nother:{}'.format(
                                     self.varnames, other.varnames))
        assert isinstance(other, type(self))
        for ch1, ch2 in zip(self.chains, other.chains):
            for k,v in ch1.items():
                allclose_kw['err_msg'] = 'Failed on {}'.format(k)
                if ignore_shape:
                    A = [np.asarray(item).flatten() for item in v]
                    B = [np.asarray(item).flatten() for item in ch2[k]]
                elif squeeze:
                    A = [np.squeeze(item) for item in v]
                    B = [np.squeeze(item) for item in ch2[k]]
                else:
                    A = v
                    B = ch2[k]
                np.testing.assert_allclose(A,B,**allclose_kw)


    ###################
    # IO and Exchange #
    ###################

    def to_df(self):
        """
        Convert the trace object to a Pandas Dataframe.

        Returns
        -------
        a dataframe where each column is a parameter. Multivariate parameters are vectorized and stuffed into a column.
        """
        dfs = []
        outnames = self.varnames
        to_split = [name for name in outnames if np.asarray(self[0,name,0]).size > 1]
        for chain in self.chains:
            out = OrderedDict(list(chain.items()))
            for split in to_split:
                records = np.asarray(copy.deepcopy(chain[split]))
                if len(records.shape) == 1:
                    records = records.reshape(-1,1)
                n,k = records.shape[0:2]
                rest = records.shape[2:]
                if len(rest) == 0:
                    pass
                elif len(rest) == 1:
                    records = records.reshape(n,int(k*rest[0]))
                else:
                    raise Exception("Parameter '{}' has too many dimensions"
                                    " to flatten able to be flattend?"               .format(split))
                records = OrderedDict([(split+'_'+str(i),record.T.tolist())
                                        for i,record in enumerate(records.T)])
                out.update(records)
                del out[split]
            df = pd.DataFrame().from_dict(out)
            dfs.append(df)
        if len(dfs) == 1:
            return dfs[0]
        else:
            return dfs

    def to_csv(self, filename, **pandas_kwargs):
        """
        Write trace out to file, going through Trace.to_df()

        If there are multiple chains in this trace, this will write
        them each out to 'filename_number.csv', where `number` is the
            number of the trace.

        Arguments
        ---------
        filename    :   string
                        name of file to write the trace to.
        pandas_kwargs:  keyword arguments
                        arguments to pass to the pandas to_csv function.
        """
        if 'index' not in pandas_kwargs:
            pandas_kwargs['index'] = False
        dfs = self.to_df()
        if isinstance(dfs, list):
            name, ext = os.path.splitext(filename)
            for i, df in enumerate(dfs):
                df.to_csv(name + '_' + str(i) + ext, **pandas_kwargs)
        else:
            dfs.to_csv(filename, **pandas_kwargs)

    @classmethod
    def from_df(cls, dfs, varnames=None, combine_suffix='_'):
        """
        Convert a dataframe into a trace object.

        Arguments
        ----------
        dfs     :   dataframe or list of dataframes
                    pandas dataframes to convert into a trace. Each dataframe is assumed to be a single chain.
        varnames:   string or list of strings
                    names to use instead of the names in the dataframe. If none, the column
                    names are split using `combine_suffix`, and the unique things before the suffix are used as parameter names.
        """
        if not isinstance(dfs, (tuple, list)):
            dfs = (dfs,)
        if len(dfs) > 1:
            traces = ([cls.from_df(df, varnames=varnames,
                        combine_suffix=combine_suffix) for df in dfs])
            return cls(*[trace.chains[0] for trace in traces])
        else:
            df = dfs[0]
        if varnames is None:
            varnames = df.columns
        unique_stems = set()
        for col in varnames:
            suffix_split = col.split(combine_suffix)
            if suffix_split[0] == col:
                unique_stems.update([col])
            else:
                unique_stems.update(['_'.join(suffix_split[:-1])])
        out = dict()
        for stem in unique_stems:
            cols = []
            for var in df.columns:
                if var == stem:
                    cols.append(var)
                elif '_'.join(var.split('_')[:-1]) == stem:
                    cols.append(var)
            if len(cols) == 1:
                targets = df[cols].values.flatten().tolist()
            else:
                # ensure the tail ordinate sorts the columns, not string order
                # '1','11','2' will corrupt the trace
                order = [int(st.split(combine_suffix)[-1]) for st in cols]
                cols = np.asarray(cols)[np.argsort(order)]
                targets = [vec for vec in df[cols].values]
            out.update({stem:targets})
        return cls(**out)

    @classmethod
    def from_pymc3(cls, pymc3trace):
        """
        Convert a PyMC3 trace to a spvcm trace
        """
        try:
            from pymc3 import trace_to_dataframe
        except ImportError:
            raise ImportError("The 'trace_to_dataframe' function in "
                              "pymc3 is used for this feature. Pymc3 "
                              "failed to import.")
        return cls.from_df(mc.trace_to_dataframe(pymc3trace))

    @classmethod
    def from_csv(cls, filename=None, multi=False,
                      varnames=None, combine_suffix='_', **pandas_kwargs):
        """
        Read a CSV into a trace object, by way of `Trace.from_df()`

        Arguments
        ----------
        filename    :   string
                        string containing the name of the file to read.
        multi       :   bool
                        flag denoting whether the trace being read is a multitrace or not. If so, the filename is understood to be the prefix of many files that end in `filename_#.csv`
        varnames    :   string or list of strings
                        custom names to use for the trace. If not provided, combine suffix is used to identify the unique prefixes in the csvs.
        pandas_kawrgs:  keyword arguments
                        keyword arguments to pass to the pandas functions.
        """
        if multi:
            filepath = os.path.dirname(os.path.abspath(filename))
            filestem = os.path.basename(filename)
            targets = [f for f in os.listdir(filepath)
                         if f.startswith(filestem)]
            ordinates = [int(os.path.splitext(fname)[0].split(combine_suffix)[-1])
                         for fname in targets]
            # preserve the order of the trailing ordinates
            targets = np.asarray(targets)[np.argsort(ordinates)].tolist()
            traces = ([cls.from_csv(filename=os.path.join(filepath, f)
                                    ,multi=False) for f in targets])
            if traces == []:
                raise IOError("No such file or directory: " +
                                        filepath + filestem)

            return cls(*[trace.chains[0] for trace in traces])
        else:
            df = pd.read_csv(filename, **pandas_kwargs)
            return cls.from_df(df, varnames=varnames,
                               combine_suffix=combine_suffix)


####################
# HELPER FUNCTIONS #
####################

def _ifilter(filt,iterable):
    """
    Filter an iterable by whether or not each item is in the filt
    """
    try:
        return iterable[filt]
    except:
        if isinstance(filt, (int, float)):
            filt = [filt]
        return [val for i,val in enumerate(iterable) if i in filt]

def _maybe_hashmap(*collections):
    """
    Attempt to coerce a collection into a Hashmap. Otherwise, leave it alone.
    """
    out = []
    for collection in collections:
        if isinstance(collection, Hashmap):
            out.append(collection)
        else:
            out.append(Hashmap(**collection))
    return out

def _copy_hashmaps(*hashmaps):
    """
    Create deep copies of the hashmaps passed to the function.
    """
    return [Hashmap(**{k:copy.deepcopy(v) for k,v in hashmap.items()})
            for hashmap in hashmaps]



from spvcm.svc import SVC 
from spvcm._constants import TEST_SEED
from spvcm.utils import baltim
import numpy as np
import os

FULL_PATH = os.path.dirname(os.path.abspath(__file__))


def build():
    inputs = baltim()
    inputs.update(dict(configs=dict(jump=.5)))
    model = SVC(**inputs, n_samples=0)
    np.random.seed(TEST_SEED)
    model.sample(1)
    model.trace.to_csv(FULL_PATH + '/data/svc.csv')


if __name__ == '__main__':
    build()



import sqlite3 as sql
import numpy as np
import os
import sys
import warnings
import pickle as pkl

LEGACY_PYTHON = sys.version_info[0] < 3

CREATE_TEMPLATE = "CREATE TABLE {} (iteration INTEGER PRIMARY KEY, {})"
INSERT_TEMPLATE = "INSERT INTO {} VALUES (?, {})"

byte_type = str if LEGACY_PYTHON else bytes

def customize_create_template(colnames, tablename):
    """
    Change the CREATE TABLE statement to bind a set of column names and table name
    """
    return CREATE_TEMPLATE.format(tablename, ' , '.join(['"' + param + '"' + ' BLOB'
                                            for param in colnames]))
def customize_insert_template(colnames, tablename):
    """
    Change the INSERT INTO statement to bind to a set of column names and table name
    """
    data_part = ' , '.join(['?' for _ in colnames])
    return INSERT_TEMPLATE.format(tablename, data_part)

def start_sql(model, tracename='model_trace.db'):
    """
    Start a SQLite connection to a local database, specified by tracename.
    """
    if os.path.isfile(tracename):
        raise Exception('Will not overwrite existing trace {}'.format(tracename))
    cxn = sql.connect(tracename)
    cursor = cxn.cursor()
    cursor.execute(customize_create_template(model.traced_params, 'trace'))
    return cxn, cursor
    
def head_to_sql(model, cursor, connection):
    """
    Send the most recent trace point to the sql database.
    """
    point_to_sql(model, cursor, connection, index=-1)

def point_to_sql(model, cursor, connection, index=0):
    """
    Send an arbitrary index point to the database
    """
    if index < 0:
        iteration = model.cycles - (index + 1)
    else:
        iteration = index
    ordered_point = (serialize(model.trace[param, index]) for param in model.traced_params)
    to_insert = [iteration]
    to_insert.extend(list(ordered_point))
    cursor.execute(customize_insert_template(model.traced_params, 'trace'), tuple(to_insert))
    connection.commit()

def trace_to_sql(model, cursor, connection):
    """
    Send a model's entire trace to the database
    """
    for i in range(model.cycles):
        ordered_point = (serialize(model.trace[param, i]) for param in model.traced_params)
        to_insert = [i]
        to_insert.extend(list(ordered_point))
        cursor.execute(customize_insert_template(model.traced_params, 'trace'), tuple(to_insert))
    connection.commit()
    
def trace_from_sql(filename, table='trace'):
    """
    Reconstruct a model trace from the database
    """
    #connect, parse header, setup trace object, then deserialize the sql
    cxn = sql.connect(filename)
    pragma = cxn.execute('PRAGMA table_info({})'.format(table)).fetchall()
    colnames = [t[1] for t in pragma]
    data = cxn.execute('SELECT * FROM {}'.format(table)).fetchall()
    cxn.close()
    records = zip(colnames, map(list, zip(*data)))

    # Import must occur here otherwise there's a circularity issue

    from .abstracts import Trace

    if table == 'trace':
        out = Trace(**{colname:[maybe_deserialize(entry) for entry in column]
                  for colname, column in records})
    else:
        out = Trace(**{colname:maybe_deserialize(column[0])
                  for colname, column in records})
    return out

def model_to_sql(model, cursor, connection):
    """
    Serialize an entire model into a sqlite database. This serializes the trace
    into the `trace` table, the state into the `state` table, and the model class 
    into the `model` table. All items are pickled using their own dumps method, if possible. 
    Otherwise, objects are reduced using dill.dumps, which is then passed to sqlite as a BLOB
    """
    trace_to_sql(model, cursor, connection)
    frozen_state_keys = list(model.state.varnames)
    frozen_state = (serialize(model.state[k]) for k in frozen_state_keys)
    cursor.execute(customize_create_template(frozen_state_keys, 'state'))
    insert_template = customize_insert_template(frozen_state_keys, 'state')
    to_insert = [model.cycles]
    to_insert.extend(list(frozen_state))
    cursor.execute(insert_template, tuple(to_insert)) 
    class_pkl = pkl.dumps(model.__class__) #want instance, not whole model
    cursor.execute(customize_create_template(['model_class'], 'model'))
    cursor.execute(customize_insert_template(['class'], 'model'), (None, class_pkl))
    connection.commit()

def model_from_sql(filename):
    """
    Reconstruct a model from a sqlite table with a given trace, state, and model tables. 

    If the serialization fails for the trace or state, the resulting
    trace/state may contain raw binary strings. If the serialization fails for the model/there
    is no model table, the function will fail. To just extract the trace or the state,
    use trace_from_sql.
    """
    trace = trace_from_sql(filename)
    state = trace_from_sql(filename, table='state')
    cxn = sql.connect(filename)
    model_class = cxn.execute('SELECT model_class FROM model')
    model_class = pkl.loads(model_class.fetchall()[0][0])
    try:
        model_class(**state)
    except:
        warn('initializing model {} from state failed! '
             'Returning trace, state, model.'.format(model_class), stacklevel=2)
        return model_class, trace, state

def maybe_deserialize(maybe_bytestring):
    """
    This attempts to deserialize an object, but may return the original object 
    if no deserialization is successful. 
    """
    try:
        import dill
    except ImportError as E:
        msg = 'The `dill` module is required to use the sqlite backend fully.'
        warnings.warn(msg, stacklevel=2)
    
    if isinstance(maybe_bytestring, (list, tuple)):
        return type(maybe_bytestring)([maybe_deserialize(byte_element) 
                                        for byte_element in maybe_bytestring])
    try:
        return pkl.loads(maybe_bytestring)
    except:
        try:
            return dill.loads(maybe_bytestring)
        except:
            try:
                return float(maybe_bytestring)
            except:
                return maybe_bytestring

def serialize(v):
    """
    This serializes an object, but may return the original object if serialization 
    is not successful. 
    """
    if hasattr(v, 'dumps'):
        return v.dumps()
    elif isinstance(v, (float, int)):
        return v
    else:
        return pkl.dumps(v)



from __future__ import division
import numpy as np
import scipy.stats as stats
import copy
MAX_SLICE = 1000

#################################
# STANDALONE SAMPLING FUNCTIONS #
#################################

def inversion(pdvec, grid):
    """
    sample from a probability distribution vector, according to a grid of values

    Parameters
    -----------
    pdvec   :   np.ndarray
                a vector of point masses that must sum to one. in theory, an
                approximation of a continuous pdf
    grid    :   np.ndarray
                a vector of values over which pdvec is evaluated. This is the
                bank of discrete values against which the new value is drawn
    """
    if not np.allclose(pdvec.sum(), 1):
        pdvec = pdvec / pdvec.sum()
    cdvec = np.cumsum(pdvec)
    a = 0
    while True:
        a += 1
        rval = np.random.random()
        topidx = np.sum(cdvec <= rval) -1
        if topidx >= 0:
            return grid[topidx]

def metropolis(state, current, proposal, logp, jump):
    """
    Sample using metropolis hastings in its simplest form

    Parameters
    ----------
    state   :   Hashmap
                state required to evaluate the logp of the parameter
    current :   float/int
                current value of the parameter
    proposal:   scipy random distribution
                distribution that has both `logpdf` and `rvs` methods
    logp    :   callable(state, value)
                function that can compute the current log of the probability
                distribution of the value provided conditional on the state
    jump    :   int
                the scaling factor for the proposal distribution
    Returns
    --------
    new (or current) parameter value, and boolean indicating whether or not a
    new proposal was accepted.
    """
    #try:
    current_logp = logp(state, current)
    new_val = proposal.rvs(loc=current, scale=jump)
    new_logp = logp(state, new_val)
    forwards = proposal.logpdf(new_val, loc=current, scale=jump)
    backward = proposal.logpdf(current, loc=new_val, scale=jump)

    hastings_factor = backward - forwards
    r = new_logp - current_logp + hastings_factor

    r = np.min((0, r))
    u = np.log(np.random.random())
    #except:
    #    print(current)
    #    print(new_val)
    #    print(new_logp)
    #    print(current_logp)
    #    print(backward)
    #    print(forwards)
    #    print(r)
    #    print(u)
    #    raise

    if u < r:
        outval = new_val
        accepted = True
    else:
        outval = current
        accepted = False
    return outval, accepted

def slicer(state, current, logp, width, adapt=10):
    """
    Implements slice sampling on a bounded log-concave parameter, as according to Neal 2003.

    Arguments
    ---------
    state   :   Hashmap/dict or tuple/list/iterable
                a collection of attributes/arguments required by the ``
                that are not `current`, the current value of the parameter
                being sampled. Is passed as a single argument to `logp`.
    current :   Float/Int
                current value of the parameter being sampled.
                NOTE: This sampler is a univariate slice sampler.
    logp    :   Callable
                function that takes (`state`, `current`) and provides the log
                probability density function to be sampled.
    width   :   float/int
                width of the slice to use when sampling.
    adapt   :   int
                weight to place on initial slice widths. Computes the moving average of the past `adapt` widths and the current width.

    """
    current_logp = logp(state, current)
    n_iterations = 0
    # p. 712 of Neal defines this auxiliary variable on the log scale
    slice_height = current_logp - np.random.exponential()

    left = current - np.random.uniform(0,width)
    right = left + width

    while slice_height < logp(state, left):
        left -= .1*width
    while slice_height < logp(state, right):
        right += .1*width


    while True:
        candidate = np.random.uniform(left, right)
        cand_logp = logp(state, candidate)
        if slice_height <= cand_logp:
            break
        if candidate > current:
            right = candidate
        elif candidate < current:
            left = candidate
        candidate = np.random.uniform(left, right)
        n_iterations += 1
        if n_iterations > MAX_SLICE:
            warn('Slicing is failing to find an effective candidate. '
                 'Using a metropolis update.', stacklevel=2)
            return metropolis(state, current, logp, configs)

    if adapt > 0:
        MA = [np.abs(left - right)]
        MA.extend([width for _ in range(adapt)])
        width = np.mean(MA)

    return candidate, True, width

def hmc(state, current_value, current_momentum, logp, dlogp):
    raise NotImplementedError

def ars(state, current_value):
    raise NotImplementedError

###############################
# SAMPLING CLASS DECLARATIONS #
###############################

class AbstractStep(object):
    """
    Standin for an abstract step
    """
    def __init__(self, varname):
        super(AbstractStep, self).__init__()
        self.varname = varname

    @property
    def _idempotent(self):
        return False

    def __call__(self, state):
        raise NotImplementedError

    def __draw__(self, state):
        return self(state)

class Gibbs(AbstractStep):
    """
    Sample directly from a given conditional log posterior distribution using the given kernel. Currently unused.
    """
    def __init__(self, varname, kernel):
        super(Gibbs, self).__init__(varname)
        self.kernel = kernel

    @property
    def _idempotent(self):
        return False

    def __call__(self, state):
        return self.kernel(state)

class Metropolis(AbstractStep):
    """
    Sample the given Logp using metroplis sampling

    Arguments
    ---------
    varname :   string
                name of variable to compute.
    logp    :   function
                function that, takes two arguments, value and state, and returns the log of the probability distribution being sampled.
    jump    :   float
                scale parameter to use in the metropolis proposal distribution
    proposal:   scipy random distribution
                distribution that has both `logpdf` and `rvs` methods
    adapt_step: float > 1
                the rate at which to adjust the jump when using tuned metropolis.
    ar_low  :   float in [0,1]
                lower bound on the acceptance rate of the metropolis sampler
    ar_hi   :   float in [0,1]
                upper bound on acceptance rate of the metrpolis sampler.
    max_tuning: int
                maximum number of iterations to tune the sampler.
    debug   :   bool
                flag denoting whether to store parameters in each iteration in the _cache.
    """
    def __init__(self, varname, logp, jump = 1, proposal = stats.norm,
                       adapt_step=1.01, ar_low = .2, ar_hi = .25, max_tuning = 0, debug=False):
        super(Metropolis, self).__init__(varname)
        self.jump = jump
        self.adapt_step = adapt_step if adapt_step >= 1 else 1/adapt_step
        self.ar_low = ar_low
        self.ar_hi = ar_hi
        self.max_tuning = max_tuning
        self.proposal = proposal
        self.accepted = 0
        self.cycles = 0
        self.logp = logp
        self.debug = debug
        if debug:
            self._cache = []

    @property
    def _idempotent(self):
        return True

    def __call__(self, state):
        orig_val = copy.deepcopy(getattr(state, self.varname))
        new, accepted = metropolis(state, orig_val,
                                    self.proposal, self.logp, self.jump)
        self.accepted += int(accepted)
        self.cycles += 1
        self.is_tuning = self.cycles <= self.max_tuning
        if self.is_tuning:
            acceptance_rate = self.accepted / self.cycles
            if acceptance_rate < self.ar_low:
                self.jump /= self.adapt_step
            if acceptance_rate > self.ar_hi:
                self.jump *= self.adapt_step
        if self.debug:
            self._cache.append(dict(jump = self.jump,
                                    current_logp = self.logp(state, orig_val),
                                    new_logp = self.logp(state, new),
                                    accepted = accepted))
        return new

class Slice(AbstractStep):
    """
    Sample the given Logp using slice sampling, of Neal (2003).

    Arguments
    ------------
    varname :   string
                name of variable to compute.
    logp    :   function
                function that, takes two arguments, value and state, and returns the log of the probability distribution being sampled.
    width   :   float
                width of the level set to use.
    adapt   :   int
                number of iterations previous to use in averaging the slice width.
    debug   :   bool
                whether or not to save the parameters in each step to the _cache attribute.
    """
    def __init__(self, varname, logp, width=1, adapt=0, debug=False):
        super(Slice, self).__init__(varname)
        self.logp = logp
        self.width = width
        self.adapt = adapt
        self.debug = debug
        if debug:
            self._cache = []

    @property
    def _idempotent(self):
        return True

    def __call__(self, state):
        new, _, width = slicer(state, getattr(state, self.varname),
                               self.logp, self.width, self.adapt)
        self.width = width
        if self.debug:
            self._cache.append(dict(width=self.width,
                                    logp = self.logp(state, new)))
        return new



__version__ = "0.3.0"
# __version__ has to be defined in the first line

"""
:mod:`spvcm` --- Spatially correlated variance components models
================================================================
"""

from .both_levels import *
from .upper_level import *
from .lower_level import *
from . import plotting



from spvcm import upper_level as upper
from spvcm import utils
from spvcm.tests.utils import Model_Mixin
from spvcm.abstracts import Trace
import unittest as ut
import pandas as pd
import os

FULL_PATH = os.path.dirname(os.path.abspath(__file__))

class Test_Upper_SE(ut.TestCase, Model_Mixin):
    def setUp(self):
        super(Test_Upper_SE, self).build_self()
        self.cls = upper.Upper_SE
        del self.inputs["W"]
        self.inputs['n_samples'] = 0
        instance = self.cls(**self.inputs)
        self.answer_trace = Trace.from_csv(FULL_PATH + '/data/upper_se.csv')



from __future__ import division

import numpy as np
import scipy.stats as stats
import copy

from numpy import linalg as la
from warnings import warn as Warn
from .sample import logp_rho_cov, logp_lambda_cov
from ...abstracts import Sampler_Mixin, Hashmap, Trace
from ... import verify
from ... steps import Metropolis, Slice
from ... import priors
from ...utils import speigen_range, ind_covariance, chol_mvn
from spreg.utils import spdot

SAMPLERS = ['Alphas', 'Betas', 'Sigma2', 'Tau2', 'Lambda', 'Rho']

class Base_Generic(Sampler_Mixin):
    """
    This is a class that provides the generic structures required for the two-level variance components models with spatially-correlated error. All of the arguments are documented in the Generic class.
    """
    def __init__(self, Y, X, W, M, Delta,
                 n_samples=1000, n_jobs=1,
                 extra_traced_params = None,
                 priors=None,
                 starting_values=None,
                 configs=None,
                 truncation=None):
        super(Base_Generic, self).__init__()

        N, p = X.shape
        _N, J = Delta.shape
        self.state = Hashmap(**{'X':X, 'Y':Y, 'M':M, 'W':W, 'Delta':Delta,
                           'N':N, 'J':J, 'p':p })
        self.traced_params = copy.deepcopy(SAMPLERS)
        if extra_traced_params is not None:
            self.traced_params.extend(extra_traced_params)
        hashmaps = [{k:[] for k in self.traced_params}]*n_jobs
        self.trace = Trace(*hashmaps)

        if priors is None:
            priors = dict()
        if starting_values is None:
            starting_values = dict()
        if configs is None:
            configs = dict()
        if truncation is None:
            truncation = dict()

        self._setup_priors(**priors)
        self._setup_configs(**configs)
        self._setup_truncation(**truncation)
        self._setup_starting_values(**starting_values)

        ## Covariance, computing the starting values
        self.state.Psi_1 = ind_covariance
        self.state.Psi_1i = ind_covariance
        self.state.Psi_2 = ind_covariance
        self.state.Psi_2i = ind_covariance

        self.cycles = 0

        if n_samples > 0:
            try:
                self.sample(n_samples, n_jobs=n_jobs)
            except (np.linalg.LinAlgError, ValueError) as e:
                Warn('Encountered the following LinAlgError. '
                     'Model will return for debugging. \n {}'.format(e))

    def _setup_priors(self, Betas_cov0 = None, Betas_mean0=None,
                      Sigma2_a0 = .001, Sigma2_b0 = .001,
                      Tau2_a0 = .001, Tau2_b0 = .001,
                      Log_Lambda0 = priors.constant,
                      Log_Rho0 = priors.constant):
        """
        This sets up the data and hyperparameters of the problem.
        If the hyperparameters are to be adjusted, pass them as keyword arguments.
        """
        st = self.state
        st.Sigma2_a0 = Sigma2_a0
        st.Sigma2_b0 = Sigma2_b0
        if Betas_cov0 is None:
            Betas_cov0 = np.eye(self.state.p) * 100
        if Betas_mean0 is None:
            Betas_mean0 = np.zeros((self.state.p, 1))
        st.Betas_cov0 = Betas_cov0
        st.Betas_mean0 = Betas_mean0
        st.Tau2_a0 = .001
        st.Tau2_b0 = .001
        st.Log_Lambda0 = Log_Lambda0
        st.Log_Rho0 = Log_Rho0

    def _finalize(self):
        """
        This computes derived properties of hyperparameters that do not change
        over iterations. This is called one time before sampling.
        """
        st = self.state

        st.In = np.eye(st.N)
        st.Ij = np.eye(st.J)

        ## Derived factors from the prior
        st.Betas_cov0i = np.linalg.inv(st.Betas_cov0)
        st.Betas_covm = np.dot(st.Betas_cov0, st.Betas_mean0)
        st.Sigma2_an = self.state.N / 2 + st.Sigma2_a0
        st.Tau2_an = self.state.J / 2 + st.Tau2_a0

        st.PsiRhoi = st.Psi_1i(st.Rho, st.W)
        st.PsiLambdai = st.Psi_2i(st.Lambda, st.M)

        ## Data invariants
        st.XtX = np.dot(self.state.X.T, self.state.X)
        st.DeltatDelta = np.dot(self.state.Delta.T, self.state.Delta)

        st.DeltaAlphas = np.dot(st.Delta, st.Alphas)
        st.XBetas = np.dot(st.X, st.Betas)

        st.initial_values = Hashmap(Betas = st.Betas, Alphas=st.Alphas, Rho=st.Rho, Lambda=st.Lambda,
                                    Sigma2=st.Sigma2, Tau2=st.Tau2)


    def _setup_configs(self, Lambda_method = 'met', Lambda_configs = None,
                             Rho_method = 'met', Rho_configs = None,
                             **uncaught):
        """
        Omnibus function to assign configuration parameters to the correct
        configuration namespace. Any options in uncaught get passed to the step methods during initialization.
        """
        if Lambda_configs is None:
            Lambda_configs = dict()
        if Rho_configs is None:
            Rho_configs = dict()

        if uncaught != dict():
            if 'method' in uncaught:
                Lambda_method = Rho_method = method
            Lambda_configs = copy.deepcopy(uncaught)
            Rho_configs = copy.deepcopy(uncaught)

        if Lambda_method.lower().startswith('met'):
            Lambda_method = Metropolis
            Lambda_configs['jump'] = Lambda_configs.pop('jump', .5)
            Lambda_configs['max_tuning'] = Lambda_configs.pop('tuning',0)
        elif Lambda_method.lower().startswith('slice'):
            Lambda_method = Slice
            Lambda_configs['width'] = Lambda_configs.pop('width', .5)
        else:
            raise Exception('Sample method for Lambda not understood:\n{}'
                            .format(Lambda_method))
        if Rho_method.lower().startswith('met'):
            Rho_method = Metropolis
            Rho_configs['jump'] = Rho_configs.pop('jump', .5)
            Rho_configs['max_tuning'] = Rho_configs.pop('tuning',0)

        elif Rho_method.lower().startswith('slice'):
            Rho_method = Slice
            Rho_configs['width'] = Rho_configs.pop('width', .5)
        else:
            raise Exception('Sample method for Rho not understood:\n{}'
                            .format(Rho_method))

        self.configs = Hashmap()
        self.configs.Rho = Rho_method('Rho', logp_rho_cov, **Rho_configs)
        self.configs.Lambda = Lambda_method('Lambda', logp_lambda_cov, **Lambda_configs)

    def _setup_truncation(self, Rho_min=None, Rho_max = None,
                          Lambda_min = None, Lambda_max = None):
        """
        This computes truncations for the spatial parameters.

        If configs.truncate is set to 'eigs', computes the eigenrange of the two
        spatial weights matrices using speigen_range

        If configs.truncate is set to 'stable', sets the truncation to -1,1

        If a tuple is passed to truncate, then this will truncate the
        distribution according to this tuple
        """
        st = self.state
        if hasattr(st, 'W'):
            if (Rho_min is None) or (Rho_max is None):
                W_emin, W_emax = speigen_range(st.W)
            if (Rho_min is None):
                Rho_min = 1./W_emin
            if (Rho_max is None):
                Rho_max = 1./W_emax
            st.Rho_min = Rho_min
            st.Rho_max = Rho_max
        if hasattr(st, 'M'):
            if (Lambda_min is None) or (Lambda_max is None):
                M_emin, M_emax = speigen_range(st.M)
            if (Lambda_min is None):
                Lambda_min = 1./M_emin
            if (Lambda_max is None):
                Lambda_max = 1./M_emax
            st.Lambda_min = Lambda_min
            st.Lambda_max = Lambda_max

    def _setup_starting_values(self, Betas = None, Alphas = None,
                               Sigma2 = 4, Tau2 = 4,
                               Rho = None, Lambda = None):
        """
        Set arbitrary starting values for the Metropolis sampler
        """
        st = self.state
        if Betas is None:
            Betas = np.zeros((self.state.p, 1))
        if Alphas is None:
            Alphas = np.zeros((self.state.J, 1))
        st.Betas = Betas
        st.Alphas = Alphas
        st.Sigma2 = Tau2
        st.Tau2 = Sigma2
        if Rho is None:
            Rho = -1.0 / (self.state.N - 1)
        if Lambda is None:
            Lambda = -1.0 / (self.state.J - 1)
        st.Rho = Rho
        st.Lambda = Lambda

    def _iteration(self):
        """
        Compute a single iteration of the sampler.
        This steps through all parameter updates exactly once.
        """
        st = self.state

        ### Sample the Beta conditional posterior
        ### P(beta | . ) \propto L(Y|.) \dot P(\beta)
        ### is
        ### N(Sb, S) where
        ### S = (X' Sigma^{-1}_Y X + S_0^{-1})^{-1}
        ### b = X' Sigma^{-1}_Y (Y - Delta Alphas) + S^{-1}\mu_0
        covm_update = spdot(st.X.T, spdot(st.PsiRhoi, st.X)) / st.Sigma2
        covm_update += st.Betas_cov0i
        covm_update = np.asarray(covm_update)
        covm_update = la.inv(covm_update)

        resids = st.Y - st.Delta.dot(st.Alphas)
        XtSresids = spdot(st.X.T, spdot(st.PsiRhoi, resids)) / st.Sigma2
        mean_update = XtSresids + st.Betas_cov0i.dot(st.Betas_mean0)
        mean_update = np.asarray(mean_update)
        mean_update = np.dot(covm_update, mean_update)
        st.Betas = chol_mvn(mean_update, covm_update)
        st.XBetas = np.dot(st.X, st.Betas)

        ### Sample the Random Effect conditional posterior
        ### P( Alpha | . ) \propto L(Y|.) \dot P(Alpha | \lambda, Tau2)
        ###                               \dot P(Tau2) \dot P(\lambda)
        ### is
        ### N(Sb, S)
        ### Where
        ### S = (Delta'Sigma_Y^{-1}Delta + Sigma_Alpha^{-1})^{-1}
        ### b = (Delta'Sigma_Y^{-1}(Y - X\beta) + 0)
        covm_update = spdot(st.Delta.T, spdot(st.PsiRhoi, st.Delta)) / st.Sigma2
        covm_update += st.PsiLambdai / st.Tau2
        covm_update = np.asarray(covm_update)
        covm_update = la.inv(covm_update)

        resids = st.Y - st.XBetas
        mean_update = spdot(st.Delta.T, spdot(st.PsiRhoi, resids)) / st.Sigma2
        mean_update = np.asarray(mean_update)
        mean_update = np.dot(covm_update, mean_update)
        st.Alphas = chol_mvn(mean_update, covm_update)
        st.DeltaAlphas = np.dot(st.Delta, st.Alphas)

        ### Sample the Random Effect aspatial variance parameter
        ### P(Tau2 | .) \propto L(Y|.) \dot P(\Alpha | \lambda, Tau2)
        ###                            \dot P(Tau2) \dot P(\lambda)
        ### is
        ### IG(J/2 + a0, u'(\Psi(\lambda))^{-1}u * .5 + b0)
        bn = spdot(st.Alphas.T, spdot(st.PsiLambdai, st.Alphas)) * .5 + st.Tau2_b0
        st.Tau2 = stats.invgamma.rvs(st.Tau2_an, scale=bn)

        ### Sample the response aspatial variance parameter
        ### P(Sigma2 | . ) \propto L(Y | .) \dot P(Sigma2)
        ### is
        ### IG(N/2 + a0, eta'Psi(\rho)^{-1}eta * .5 + b0)
        ### Where eta is the linear predictor, Y - X\beta + \DeltaAlphas
        eta = st.Y - st.XBetas - st.DeltaAlphas
        bn = spdot(eta.T, spdot(st.PsiRhoi, eta)) * .5 + st.Sigma2_b0
        st.Sigma2 = stats.invgamma.rvs(st.Sigma2_an, scale=bn)

        ### Sample the spatial components using metropolis-hastings
        ### P(Psi(\lambda) | .) \propto L(Y | .) \dot P(\lambda)
        ### is
        ### |Psi(lambda)|^{-1/2} exp(1/2(Alphas'Psi(lambda)^{-1}Alphas * Tau2^{-1}))
        ###  * 1/(emax-emin)
        st.Rho = self.configs.Rho(st)

        st.PsiRhoi = st.Psi_1i(st.Rho, st.W) #BtB or CiCti

        ### P(Psi(\rho) | . ) \propto L(Y | .) \dot P(\rho)
        ### is
        ### |Psi(rho)|^{-1/2} exp(1/2(eta'Psi(rho)^{-1}eta * Sigma2^{-1})) * 1/(emax-emin)
        st.Lambda = self.configs.Lambda(st)

        st.PsiLambdai = st.Psi_2i(st.Lambda, st.M)

class Generic(Base_Generic):
    """
    This is a class that provides the generic structures required for the two-level variance components models with spatially-correlated error.

    Arguments
    ----------
    Y       :   numpy.ndarray
                The (n,1) array containing the response to be modeled
    X       :   numpy.ndarray
                The (n,p) array containing covariates used to predict the  response, Y.
    W       :   pysal.weights.W
                a spatial weights object for the n observations
    M       :   pysal.weights.W
                a spatial weights object for the J regions.
    Z       :   numpy.ndarray
                The (J,p') array of region-level covariates used to predict the response, Y.
    Delta   :   numpy.ndarray
                The (n,J) dummy variable matrix, relating observation i to region j. Required if membership is not passed.
    membership: numpy.ndarray
                The (n,) vector of labels assigning each observation to a region. Required if Delta is not passed.
    transform  : string
                 weights transform to use in modeling.
    verbose    : bool
                 whether to provide verbose output about sampling progress
    n_samples : int
                the number of samples to draw. If zero, the model will only be initialized, and later sampling can be done using model.sample
    n_jobs  :   int
                the number of parallel chains to run. Defaults to 1.
    extra_traced_param  :   list of strings
                            extra parameters in the trace to record.
    center  :   bool
                Whether to center the input data so that its mean is zero. Occasionally improves the performance of the sampler.
    scale   :   bool
                Whether to rescale the input data so that its variance is one. May dramatically improve the performance of the sampler, at the cost of interpretability of the coefficients.
    priors  :   dictionary
                A dictionary used to configure the priors for the model. This may include the following keys:
                    Betas_cov0  : prior covariance for Beta parameter vector
                                (default: I*100)
                    Betas_mean0 : prior mean for Beta parameters
                                (default: 0)
                    Sigma2_a0   : prior shape parameter for inverse gamma prior on response-level variance
                                (default: .001)
                    Sigma2_b0   : prior scale parameter for inverse gamma prior on response-level variance
                                (default: .001)
                    Tau2_a0     : prior shape parameter for inverse gamma prior on regional-level variance
                                (default: .001)
                    Tau2_b0     : prior scale parameter for inverse gamma prior on regional-level variance
                                (default: .001)
                    Log_Lambda0 : prior on Lambda, the region-level autoregressive parameter. Must be a callable function that takes a single argument and returns a single value providing the log prior likelihood.
                                (default: priors.constant)
                    Log_Rho0    : prior on Rho, the response-level autoregressive paraameter. Must be a callable function that takes a single argument and returns a single value providing the log prior likelihood.
                                (default: priors.constant)
    starting_value :    dictionary
                        A dictionary containing the starting values of the sampler. If n_jobs > 1, these starting values are perturbed randomly to induce overdispersion.

                        This dicutionary may include the following keys:
                        Betas   : starting vector of Beta parameters.
                                (default: np.zeros(p,1))
                        Alphas  : starting vector of Alpha variance components.
                                (default: np.zeros(J,1))
                        Sigma2  : starting value of response-level variance
                                (default: 4)
                        Tau2    : starting value of region-level varaince
                                (default: 4)
                        Rho     : starting value of response-level spatial autoregressive parameter
                                (default: -1/n)
                        Lambda  : starting value of region-level spatial autoregressive parameter
                                (default -1/J)
    config  :   dictionary
                A dictionary containing the configuration values for the non-Gibbs samplers for the spatial parameters. May be specified for each parameter individually (if both are in the model), or may be specified implicitly for the relevant parameter to the model. Keys may include:
                Rho_method      : string specifying whether "met" or "slice" sampling should be used for the response-level effects
                Rho_configs     : configuration options for the sampler for the response-level effects that will be used in the step method.
                Lambda_method   : string specifying whether 'met' or 'slice' sampling should be used for the region-level effects
                Lambda_configs  : configuration options for the sampler for the region-level effects that will be used in the step method.

                For options that can be in Lambda/Rho_configs, see:
                spvcm.steps.Slice, spvcm.steps.Metropolis
    truncation  :   dictionary
                    A dictionary containing the configuration values for the maximum and minimum allowable Lambda and Rho parameters. If these are not provided, the support is row-standardized by default, and the minimal eigenvalue computed for the lower bound on the parameters. *only* the single minimum eigenvalue is computed, so this is still rather efficient for large matrices. Keys may include:
                    Rho_min     : minimum value allowed for response-level autoregressive coefficient
                    Rho_max     : maximum value allowed for response-level autoregressive coefficient
                    Lambda_min  : minimum value allowed for region-level autoregressive coefficient
                    Lambda_max  : maximum value allowed for region-level autoregressive coefficient.
    """
    def __init__(self, Y, X, W, M, Z=None, Delta=None, membership=None,
                 #data options
                 transform ='r', verbose=False,
                 n_samples=1000, n_jobs=1,
                 extra_traced_params = None,
                 priors=None,
                 starting_values=None,
                 configs=None,
                 truncation=None,
                 center=False,
                 scale=False):
        if X is None:
            X = np.ones_like(Y)
            center=False
            scale=False
        W,M = verify.weights(W,M, transform=transform)
        self.M = M


        N,_ = X.shape
        J = M.n
        Mmat = M.sparse
        Wmat = W.sparse
        Delta, membership = verify.Delta_members(Delta, membership, N, J)


        if Z is not None:
            Z = Delta.dot(Z)
            X = np.hstack((X,Z))
        if center:
            X = verify.center(X)
        if scale:
            X = verify.scale(X)

        X = verify.covariates(X)

        self._verbose = verbose

        super(Generic, self).__init__(Y, X, Wmat, Mmat, Delta,
                                      n_samples=n_samples, n_jobs=n_jobs,
                                      extra_traced_params = extra_traced_params,
                                      priors=priors,
                                      starting_values=starting_values,
                                      configs=configs,
                                      truncation=truncation)



import numpy as np
from spreg.utils import sphstack
from spreg.diagnostics import constant_check
from warnings import warn as Warn

def weights(W,M, transform):
    """
    This tries to transform a pysal spatial weights object into being
    row-standardized. It warns if the objects do not support transformation.

    """
    try:
        if M is not None:
            M.transform = 'r'
        if W is not None:
            W.transform = 'r'
    except AttributeError:
        Warn("Weights objects do not support transformation. Proceeding without transforming weights.", UserWarning)
    try:
        if M is not None and W is not None:
            assert M.n <= W.n
    except AssertionError:
        raise AssertionError('M (n={}) is larger than W (n={}).'.format(M.n, W.n))
    return W,M

def covariates(X):
    """
    This

    1. checks if the lower-level covariate contains a constant
    2. adds a constant to the lower-level covariate if it has no constant
    """
    if constant_check(X):
        Warn("X array should not contain a constant vector;"
             " constant will be added automatically")
    else:
        X = sphstack(np.ones((X.shape[0], 1)), X)

    return X

def center(*arrays, **kw):
    """
    This centers the arrays provided along the axis provided.
    """
    axis = kw.pop('axis', 0)
    out = [(array - array.mean(axis=axis)) for array in arrays]
    return out if len(out) > 1 else out[0]

def scale(*arrays, **kw):
    """
    This scales covariates by their standard deviation along the axis provided.
    """
    axis = kw.pop('axis', 0)
    out = [array/array.std(axis=axis) for array in arrays]
    return out if len(out) > 1 else out[0]


def Delta_members(Delta, membership, N, J):
    """
    This computes and verifies a Delta or membership vector.
    """
    if Delta is None and membership is None:
        raise UserWarning("No Delta matrix nor membership classification provided. Refusing to arbitrarily assign units to upper-level regions.")
    elif membership is None:
        membership = np.zeros((N,1))
        for idx, region in enumerate(Delta.T):
            membership[region.flatten() == 1] = idx
    elif Delta is None:
        Delta = np.zeros((N, J))
        for idx, region in enumerate(np.unique(membership)):
            Delta[membership.flatten() == region, idx] = 1
    else:
        raise UserWarning("Both Delta and Membership vector provided. Please pass only one or the other.")
    return Delta, membership



import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

__all__ = ['plot_trace', 'seplot', 'rollplot', 'corrplot']

def plot_trace(model, burn=0, thin=None, varnames=None, trace=None,
               kde_kwargs={}, trace_kwargs={}, figure_kwargs={}):
    """
    Make a trace plot paired with a distributional plot.

    Arguments
    -----------
    model   :   Model object
                a model with a trace attribute
    burn    :   int
                the number of iterations to discard from the front of the trace.
                If negative, the number of iterations to use from the tail of the trace.
    thin    :   int
                the number of iterations to discard between iterations
    varnames :  str or list
                name or list of names to plot.
    trace   :   namespace
                a namespace whose variables are contained in varnames
    kde_kwargs : dictionary
                 dictionary of aesthetic arguments for the kde plot
    trace_kwargs : dictionary
                   dictinoary of aesthetic arguments for the traceplot
    figure_kwargs: dictionary
                    a dictionary of arguments for the plot creator

    Returns
    -------
    figure, axis tuple, where axis is (len(varnames), 2)
    """
    if thin is None or thin is 0:
        thin = 1
    if model is None:
        if trace is None:
            raise Exception('Neither model nor trace provided.')
    else:
        trace = model.trace
    if varnames is None:
        varnames = trace.varnames
    elif isinstance(varnames, str):
        varnames = [varnames]
    if figure_kwargs == dict():
        figure_kwargs = {'figsize':(8, 2*len(varnames)), 'sharey':'row'}
    if kde_kwargs == dict():
        kde_kwargs = {'shade':True, 'vertical':True}
    if trace_kwargs == dict():
        trace_kwargs = {'linewidth':.5}
    fig, ax = plt.subplots(len(varnames), 2, **figure_kwargs)
    for chain_i, chain in enumerate(trace.chains):
        for i, param_name in enumerate(varnames):
            this_param = np.asarray(trace[chain_i,param_name,burn::thin])
            if ax.ndim ==1:
                ax = np.array((ax,),)
            if len(this_param.shape) == 3:
                n,a,b = this_param.shape
                this_param = this_param.reshape(n,a*b)
            if len(this_param.shape) == 2:
                if this_param.shape[-1] == 1:
                    sns.kdeplot(this_param.flatten(),
                                ax=ax[i,1], **kde_kwargs)
                else:
                    for param in this_param.T:
                        sns.kdeplot(param, ax=ax[i,1], **kde_kwargs)
            else:
                sns.kdeplot(this_param, ax=ax[i,1], **kde_kwargs)
            ax[i,0].plot(this_param, **trace_kwargs)
            ax[i,1].set_title(param_name)
            ax[i,0].set_xbound(0, len(this_param))
    fig.tight_layout()
    return fig, ax

def seplot(model=None, trace=None, chain=None, varnames=None,
           burn=0, thin=None, N_bins=200,
           plot_kw=None, fig_kw=None, ax=None):
    """
    This plots the markov chain monte carlo standard error 
    as a function of iterations. 

    Arguments
    ---------
    model   :   Any model object.
                must have an attached `trace` attribute. Takes precedence over
                `trace` and `chain` arguments.
    trace   :   abstracts.Trace
                a trace object containing data to compute the diagnostic
    chain   :   np.ndarray
                an array indexed by (m,n[,p]) containing m parallel runs of n samples
                of p covariates.
    varnames:   str or list of str
                set of variates to extract from the model or trace to to compute the
                statistic.
    N_bins  :   int
                number of subsamples over which to compute the standard error
    roller  :   callable
                pandas rolling function that takes argument `window` to specify window size
    burn    :   int
                number of iterations to discard from the front of the chain. If negative, number of iterations to keep from the tail of the chain.
    thin    :   int
                step to thin the chain. Picks every `thin`th observation.
    plot_kw :   dict/keyword arguments
                passed to the line plot call
    fig_kw  :   dict/keyword arguments
                passed to the plt.subplots() call
    ax      :   matplotlib axis
                pass if you want to plot this on an existing set of axes.

    Returns
    --------
    figure,axis tuple or, if ax is passed, ax
    """
    from . import diagnostics as diag
    trace = diag._resolve_to_trace(model, trace, chain, varnames)

    thin = 1 if thin is None else thin

    trace = diag._resolve_to_trace(None, trace[burn::thin], None, None)
    statistics = trace.map(_se_vector, N_bins=N_bins)

    if fig_kw is None:
        fig_kw = dict()
    fig_kw['figsize'] = fig_kw.get('figsize', (5, 2*len(trace.varnames)))
    fig_kw['sharex'] = fig_kw.get('sharex', True)
    if plot_kw is None:
        plot_kw = dict()

    if ax is None:
        f,ax = plt.subplots(len(trace.varnames), 1, **fig_kw)

    for i, varname in enumerate(trace.varnames):
        to_plot = (np.asarray(result[varname]) for result in statistics)
        for results in to_plot:
            if results.ndim > 1:
                results = results.T
            ax[i].plot(results, **plot_kw)
            ax[i].set_title(varname)
        ax[i].set_ylabel('MC Standard Error')
    try:
        f.tight_layout()
        return f,ax
    except NameError:
        return ax


def rollplot(model=None, trace=None, chain=None, varnames=None,
             order=100, roller=None,
             burn=0, thin=None, plot_kw=None, fig_kw=None, ax=None):
    """
    This plots a rolling window function, `roller`, against all parameters contained
    in the model, trace, or chain. 

    Arguments
    ---------
    model   :   Any model object.
                must have an attached `trace` attribute. Takes precedence over
                `trace` and `chain` arguments.
    trace   :   abstracts.Trace
                a trace object containing data to compute the diagnostic
    chain   :   np.ndarray
                an array indexed by (m,n[,p]) containing m parallel runs of n samples
                of p covariates.
    varnames:   str or list of str
                set of variates to extract from the model or trace to to compute the
                statistic.
    order   :   int
                the order of the moving process
    roller  :   callable
                function that takes arguments `data, window`, where `data` is all data and `window` is the size
                of the window to pass over `data`, applying function `roller`. 
    burn    :   int
                number of iterations to discard from the front of the chain. If negative, number of iterations to keep from the tail of the chain.
    thin    :   int
                step to thin the chain. Picks every `thin`th observation.
    plot_kw :   dict/keyword arguments
                passed to the line plot call
    fig_kw  :   dict/keyword arguments
                passed to the plt.subplots() call
    ax      :   matplotlib axis
                pass if you want to plot this on an existing set of axes.

    Returns
    --------
    figure,axis tuple or, if ax is passed, ax
    """
    from . import diagnostics as diag
    if roller is None:
        roller = lambda data, order: pd.Series(data).rolling(order).mean()

    trace = diag._resolve_to_trace(model, trace, chain, varnames)

    thin = 1 if thin is None else thin
    trace = diag._resolve_to_trace(model, trace[burn::thin], chain, varnames)

    if fig_kw is None:
        fig_kw = dict()
    fig_kw['figsize'] = fig_kw.get('figsize', (5, 2*len(trace.varnames)))
    fig_kw['sharex'] = fig_kw.get('sharex', True)
    if plot_kw is None:
        plot_kw = dict()

    if ax is None:
        f,ax = plt.subplots(len(trace.varnames), 1, **fig_kw)
    for i, varname in enumerate(trace.varnames):
        for j in range(trace.n_chains):
            data = trace[j, varname]
            ax[i].plot(roller(data, order), **plot_kw)
        ax[i].set_title(varname)
    try:
        f.tight_layout()
        return f,ax
    except NameError:
        return ax


def conv_plot(model=None, trace=None, chain=None, varnames=None,
              N_bins=200, roller=None,
              burn=0, thin=None, plot_kw=None, fig_kw=None, ax=None):
    """
    This plots a moving average of the chain alongside of a standard error indicator. This takes a long time to compute and provides poor plots unless N_bins is very high relative to the chain length.

    Arguments
    ---------
    model   :   Any model object.
                must have an attached `trace` attribute. Takes precedence over
                `trace` and `chain` arguments.
    trace   :   abstracts.Trace
                a trace object containing data to compute the diagnostic
    chain   :   np.ndarray
                an array indexed by (m,n[,p]) containing m parallel runs of n samples
                of p covariates.
    varnames:   str or list of str
                set of variates to extract from the model or trace to to compute the
                statistic.
    N_bins  :   int
                the number of sample points to use to compute the rolling mean and standard errors.
    roller  :   callable
                function that takes arguments `data, window`, where `data` is all data and `window` is the size
                of the window to pass over `data`, applying function `roller`. 
    burn    :   int
                number of iterations to discard from the front of the chain. If negative, number of iterations to keep from the tail of the chain.
    thin    :   int
                step to thin the chain. Picks every `thin`th observation.
    plot_kw :   dict/keyword arguments
                passed to the line plot call
    fig_kw  :   dict/keyword arguments
                passed to the plt.subplots() call
    ax      :   matplotlib axis
                pass if you want to plot this on an existing set of axes.

    Returns
    --------
    figure,axis tuple or, if ax is passed, ax

    """
    from . import diagnostics as diag
    if roller is None:
        roller = lambda data, order: pd.Series(data).rolling(order).mean()

    trace = diag._resolve_to_trace(model, trace, chain, varnames)

    thin = 1 if thin is None else thin

    trace = diag._resolve_to_trace(None, trace[burn::thin], None,None)

    if fig_kw is None:
        fig_kw = dict()
    fig_kw['figsize'] = fig_kw.get('figsize', (5, 2*len(trace.varnames)))
    fig_kw['sharex'] = fig_kw.get('sharex', 'col')
    if plot_kw is None:
        plot_kw = dict()

    f,ax = plt.subplots(len(trace.varnames), 1, **fig_kw)

    window = trace.n_iters - N_bins + 1
    rolled = trace.map(roller, window=window)
    ses = trace.map(_se_vector, N_bins=N_bins)

    for i, varname in enumerate(trace.varnames):
        for j in range(trace.n_chains):
            this_roll = np.asarray(rolled[j][varname])
            if this_roll.ndim > 1:
                this_roll = this_roll[~np.isnan(this_roll)].reshape(-1, N_bins)
            else:
                this_roll = this_roll[~np.isnan(this_roll)]
            this_se = np.asarray(ses[j][varname])

            upper = this_roll + this_se * 2
            lower = this_roll - this_se * 2
            if this_roll.ndim > 1:
                for this, up, low in zip(this_roll, upper, lower):
                    ax[i].fill_between(np.arange(len(this)), up, low, where = up>low, alpha=.5)
            else:
                ax[i].fill_between(np.arange(len(this_roll)), upper, lower, where = upper>lower, alpha=.5)
            ax[i].plot(this_roll, **plot_kw)
        ax[i].set_title(varname)
    return f,ax


def _se_vector(x, N_bins):
    """
    A vectorized version of the monte carlo standard error estimator. Computes the
    monte carlo standard error by breaking `x` into `N_bins`, and then compuing the standard error over the chain defined by concatenating the bins starting at the head of the chain.

    If the chain is converged, the standard error estimator should converge to zero in this procedure. The speed of convergence is indicated by how quickly the standard error reduces.

    Arguments
    ----------
    x       :   numpy.ndarray
                an array over which to compute the standard errors
    N_bins  :   int
                number of bins to break `x` into to compute standard errors

    Returns
    ----------
    N_bins-length vector containing whose element i contains the standard error of the x[:i+1] chunk.
    """
    from . import diagnostics as diag

    splits = np.array_split(x, N_bins)
    diags = [list(diag.mcse(chain=np.hstack(splits[:i+1])).values())[0] for i in range(N_bins)]
    return diags

def corrplot(m, burn=0, thin=None,
             percentiles=[25,50,75], support=np.linspace(.001,1,num=1000),
             figure_kw=None, plot_kw=None, kde_kw=None):
    if figure_kw is None:
        figure_kw = {'figsize':(2.1*8,8), 'sharey':True}

    if plot_kw is None:
        plot_kw = [dict()]*len(percentiles)
    elif isinstance(plot_kw, dict):
        plot_kw = [plot_kw]*len(percentiles)
    elif isinstance(plot_kw, list):
        assert len(plot_kw)==len(percentiles)

    if kde_kw is None:
        kde_kw = [{'vertical':True, 'shade':True}]*len(percentiles)
    elif isinstance(kde_kw, dict):
        kde_kw = [kde_kw]*len(percentiles)
    elif isinstance(kde_kw, list):
        assert len(kde_kw)==len(percentiles)

    corrfunc = m.state.correlation_function
    pwds = m.state.pwds
    if m.trace.n_chains > 1:
        raise
    phis = m.trace['Phi', burn::thin]
    f,ax = plt.subplots(1,2, **figure_kw)
    support = np.linspace(.001,1,num=1000)
    ptiles = [[np.percentile(corrfunc(r, pwds).flatten(), ptile)
               for r in support] for ptile in percentiles]
    empirical_median_correlation = [np.median(corrfunc(phi, pwds)) for phi in phis]
    for i, ptile in enumerate(ptiles):
        ax[0].plot(support*m.state.max_dist, ptile, **plot_kw[i])
    sns.kdeplot(np.asarray(empirical_median_correlation), ax=ax[1], **kde_kw[i])
    ax[0].set_title('Percentile of Correlation')
    ax[1].set_title('Median spatial correlations')
    ax[1].set
    ax[0].set_ybound(0,1)
    ax[0].set_xlabel('Distance')
    ax[0].set_ylabel('Inter-Observation $\\rho$')
    return f, ax


def hpd_trajplot(model=None, trace=None, chain=None, varnames=None,
                 alpha=.95, n_splits=100,
                 fig_kw=dict(), hpdi_kw=dict(), trace_kw=dict(), width_kw=dict()):
    """
    This plots the middle alpha% interval around the parameter over n_splits blocks. 

     ---------
    model   :   Any model object.
                must have an attached `trace` attribute. Takes precedence over
                `trace` and `chain` arguments.
    trace   :   abstracts.Trace
                a trace object containing data to compute the diagnostic
    chain   :   np.ndarray
                an array indexed by (m,n[,p]) containing m parallel runs of n samples
                of p covariates.
    varnames:   str or list of str
                set of variates to extract from the model or trace to to compute the
                statistic.
    alpha   :   float
                the percentile to use for the HPD interval plotted in each chunk
    n_splits:   int
                the number of chunks of the chain to estimate the HPD over. 
    burn    :   int
                number of iterations to discard from the front of the chain. If negative,
                number of iterations to keep from the tail of the chain.
    thin    :   int
                step to thin the chain. Picks every `thin`th observation.
    fig_kw  :   dict/keyword arguments
                passed to the plt.subplots() call
    hpdi_kw :   dict/keyword arguments
                passed to the diagnostics.hpd_interval() call for the hpd envelope
    trace_kw:   dict/keyword arguments
                passed to the plt.plot() call for the trace of the parameters
    width_kw:   dict/keyword arguments
                passed to the plt.plot() call for the intervals of parameters
    ax      :   matplotlib axis
                pass if you want to plot this on an existing set of axes.

    Returns
    --------
    figure,axis tuple or, if ax is passed, ax

    """
    from . import diagnostics as diag
    trace = diag._resolve_to_trace(model, trace, chain, varnames)
    if varnames is None:
        varnames = trace.varnames
    p = len(varnames)
    f, ax = plt.subplots(p, 2, **fig_kw)
    pieces = trace.map(np.array_split, indices_or_sections=n_splits)
    hpds = dict()
    for i, varname in enumerate(varnames):
        bits = pieces[varname]
        hpds[varname] = []
        for i, bit in enumerate(bits):
            try:
                cumulant = np.hstack((cumulant, bit))
                hpd = diag.hpd_interval(chain=cumulant, alpha=alpha)
            except NameError:
                cumulant = bit
                hpd = diag.hpd_interval(chain=cumulant, alpha=alpha)
            finally:
                hpds[varname].append(hpd)
        this_hpdset = np.hstack(hpds[varname]).reshape(-1, 2)
        keff = len(this_hpdset)
        support = np.arange(0, keff)
        i, ax[0].plot(support, this_hpdset.T[0], **hpdi_kw)
        i, ax[0].plot(support, this_hpdset.T[1], **hpdi_kw)
        i, ax[0].plot(trace[varname], **trace_kw)
        widths = np.subtract.reduce(this_hpdset, axis=1)
        i, ax[1].plot(support, widths, **width_kw)
    return f, ax



import os as _os

def available():
    dcty = _os.path.dirname(_os.path.abspath(__file__))
    files = _os.listdir(dcty)

    unique_names = {_os.path.splitext(f)[0]:[] for f in files 
                    if not _os.path.splitext(f)[0].startswith('__init__')}

    for f in files:
        filename = _os.path.splitext(f)[0]
        if filename not in unique_names:
            continue
        unique_names[filename].append(f)
    
    print('The following examples are stored in the '
          'examples directory\n{}'.format(dcty))
    print(unique_names)



from spvcm._constants import TEST_SEED, CLASSTYPES
from spvcm.tests.utils import run_with_seed
from spvcm.api import upper
from spvcm.abstracts import Sampler_Mixin
from spvcm.utils import south
import pandas as pd
import os

FULL_PATH = os.path.dirname(os.path.abspath(__file__))

def build():
    models = []
    for cand in upper.__dict__.values():
        if isinstance(cand, CLASSTYPES):
            if issubclass(cand, Sampler_Mixin):
                models.append(cand)
    for model in models:
        env = south()
        del env['W']
        print('starting model {}'.format(model))
        run_with_seed(model, env=env, seed=TEST_SEED, fprefix=FULL_PATH + '/data/')
    return os.listdir(FULL_PATH + '/data/')



from spvcm.utils import south
import spvcm.api as spvcm
from spvcm._constants import TEST_SEED
from spvcm.diagnostics import psrf, mcse, hpd_interval, effective_size, geweke
import numpy as np
import os

FULL_PATH = os.path.dirname(os.path.abspath(__file__))


def build():
    data = south()
    del data['W']
    del data['M']
    model = spvcm.both.MVCM(**data, n_samples=0)
    np.random.seed(TEST_SEED)
    print('starting South 5000, njobs=4')
    model.sample(5000,n_jobs=4)
    print('starting PSRF')
    known_brooks = psrf(model)
    known_gr = psrf(model, method='original')


    import json
    with open(FULL_PATH + '/data/' + 'psrf_brooks.json', 'w') as brooks:
        json.dump(known_brooks, brooks)
    with open(FULL_PATH + '/data/' + 'psrf_gr.json', 'w') as gr:
        json.dump(known_gr, gr)
    for i, method in enumerate(['bm', 'obm', 'bartlett', 'hanning']):
        known_mcse = mcse(model, varnames=['Tau2'], method=method)
        with open(FULL_PATH + '/data/' + 'mcse_{}.json'.format(i,method), 'w') as mcse_file:
            json.dump(known_mcse, mcse_file)

    known_hpd = hpd_interval(model, varnames=['Sigma2'])
    with open(FULL_PATH + '/data/' + 'hpd_interval.json', 'w') as hpd_file:
        json.dump(known_hpd, hpd_file)

    known_size = effective_size(model, varnames=['Tau2'], use_R=False)
    with open(FULL_PATH + '/data/' + 'effective_size.json', 'w') as size_file:
        json.dump(known_size, size_file)

    known_geweke = geweke(model, varnames=['Sigma2'])
    known_geweke = [{k:v.tolist() for k,v in known.items()} for known in known_geweke]
    with open(FULL_PATH + '/data/' + 'geweke.json', 'w') as geweke_file:
        json.dump(known_geweke, geweke_file)

    model.trace.to_csv(FULL_PATH + '/data/' + 'south_mvcm_5000.csv')
    return ([FULL_PATH + '/data/' + 'psrf_{}.json'.format(k)
             for k in ['brooks', 'gr']] + [FULL_PATH + '/data/south_mvcm_5000.csv'])
